import re
from urllib.parse import urlparse

from .url import URL


class UrlParser(object):

    def __init__(self, url):
        self.url = self._get_valid_url(url)

    def host(self):
        return self.url.host()

    def domain(self):
        return self.url.domain()

    def path(self):
        return self.url.path()

    def fragment(self):
        return self.url.fragment()

    def fragment_split_get_first(self, params):
        return self.url.fragment().split(params, 1)[0]

    def scheme(self):
        return self.url.scheme()

    def query(self):
        return self.url.query()

    def query_param(self, params):
        return self.url.query_param(params)

    @staticmethod
    def _get_valid_url(url):
        converted_url = u'{}'.format(url)

        if not urlparse(converted_url).scheme:
            converted_url = 'http://{}'.format(converted_url)

        return URL(converted_url)

    def query_param_with_semicolon_and_not_unquote(self, params):
        return self.url.query_param_with_semicolon_and_not_unquote(params)

    def query_param_in_path(self, params):
        return self.url.query_param_in_path(params)

    def path_re_search_group(self, params):
        url_part = self.url.path()
        return self._re_search_group(params, url_part)

    def host_re_search_group(self, params):
        url_part = self.url.host()
        return self._re_search_group(params, url_part)

    def domain_re_search_group(self, params):
        url_part = self.url.domain()
        return self._re_search_group(params, url_part)

    def fragment_re_search_group(self, params):
        url_part = self.url.fragment()
        return self._re_search_group(params, url_part)

    def query_re_search_group(self, params):
        url_part = self.url.query()
        return self._re_search_group(params, url_part)

    def _re_search_group(self, params, url_part):
        group = int(params[:1])
        regex = params[2:]
        return re.compile(regex).search(url_part).group(group)



from urllib.parse import parse_qs as original_parse_qs
from urllib.parse import _coerce_args

from purl import URL as PURL_URL
from purl.url import (
    dict_to_unicode, six, to_unicode, to_utf8, unicode_urlencode, unquote
)


def parse_qsl_with_semicolon_and_not_unquote(
        qs, keep_blank_values=False, strict_parsing=False, encoding='utf-8', errors='replace'):
    """
    This function was copied and modified from https://hg.python.org/cpython/file/3.4/Lib/urllib/parse.py
    to test quickly, will probably need refactor in the future.

    Parse a query given as a string argument.

    Arguments:

    qs: percent-encoded query string to be parsed

    keep_blank_values: flag indicating whether blank values in
        percent-encoded queries should be treated as blank strings.  A
        true value indicates that blanks should be retained as blank
        strings.  The default false value indicates that blank values
        are to be ignored and treated as if they were  not included.

    strict_parsing: flag indicating what to do with parsing errors. If
        false (the default), errors are silently ignored. If true,
        errors raise a ValueError exception.

    encoding and errors: specify how to decode percent-encoded sequences
        into Unicode characters, as accepted by the bytes.decode() method.

    Returns a list, as G-d intended.
    """
    qs, _coerce_result = _coerce_args(qs)
    pairs = [s1 for s1 in qs.split('&')]
    r = []
    for name_value in pairs:
        if not name_value and not strict_parsing:
            continue
        nv = name_value.split('=', 1)
        if len(nv) != 2:
            if strict_parsing:
                raise ValueError("bad query field: %r" % (name_value,))
            # Handle case of a control-name with no equal sign
            if keep_blank_values:
                nv.append('')
            else:
                continue
        if len(nv[1]) or keep_blank_values:
            name = nv[0].replace('+', ' ')
            name = unquote(name, encoding=encoding, errors=errors)
            name = _coerce_result(name)
            value = nv[1]
            r.append((name, value))
    return r


def parse_qs_with_semicolon_and_not_unquote(
        qs, keep_blank_values=False, strict_parsing=False, encoding='utf-8', errors='replace'):
    """
    This function was copied and modified from https://hg.python.org/cpython/file/3.4/Lib/urllib/parse.py
    to test quickly, will probably need refactor in the future.

    Parse a query given as a string argument.

    Arguments:

    qs: percent-encoded query string to be parsed

    keep_blank_values: flag indicating whether blank values in
        percent-encoded queries should be treated as blank strings.
        A true value indicates that blanks should be retained as
        blank strings.  The default false value indicates that
        blank values are to be ignored and treated as if they were
        not included.

    strict_parsing: flag indicating what to do with parsing errors.
        If false (the default), errors are silently ignored.
        If true, errors raise a ValueError exception.

    encoding and errors: specify how to decode percent-encoded sequences
        into Unicode characters, as accepted by the bytes.decode() method.
    """
    parsed_result = {}
    pairs = parse_qsl_with_semicolon_and_not_unquote(
        qs, keep_blank_values, strict_parsing, encoding=encoding, errors=errors)
    for name, value in pairs:
        if name in parsed_result:
            parsed_result[name].append(value)
        else:
            parsed_result[name] = [value]
    return parsed_result


class URL(PURL_URL):

    def query_param_with_semicolon_and_not_unquote(self, key, value=None, default=None, as_list=False):
        return self._query_param(
            key, value=value, default=default, as_list=as_list,
            func=self.query_params_with_semicolon_and_not_unquote)

    def query_param_in_path(self, key, value=None, default=None, as_list=False, func=None):
        return self._query_param(
            key, value=value, default=default, as_list=as_list,
            func=self.query_params_in_path)

    def _query_param(self, key, value=None, default=None, as_list=False, func=None):
        """
        This function was copied and modified from https://github.com/codeinthehole/purl/blob/master/purl/url.py
        to test quickly, will probably need refactor in the future.

        Return or set a query parameter for the given key
        The value can be a list.
        :param string key: key to look for
        :param string default: value to return if ``key`` isn't found
        :param boolean as_list: whether to return the values as a list
        :param string value: the new query parameter to use
        """
        if func is None:
            func = self.query_params

        parse_result = func()

        if value is not None:
            # Need to ensure all strings are unicode
            if isinstance(value, (list, tuple)):
                value = list(map(to_unicode, value))
            else:
                value = to_unicode(value)
            parse_result[to_unicode(key)] = value
            return URL._mutate(
                self, query=unicode_urlencode(parse_result, doseq=True))

        try:
            result = parse_result[key]
        except KeyError:
            return default
        if as_list:
            return result
        return result[0] if len(result) == 1 else result

    def query_params_with_semicolon_and_not_unquote(self, value=None):

        query = '' if self._tuple.query is None else self._tuple.query

        return self._query_params(value=value, qs_func=parse_qs_with_semicolon_and_not_unquote, query=query)

    def query_params_in_path(self, value=None):

        path = '' if self._tuple.path is None else self._tuple.path

        return self._query_params(value=value, qs_func=original_parse_qs, query=path)

    def _query_params(self, value=None, qs_func=None, query=None):
        """
        This function was copied and modified from https://github.com/codeinthehole/purl/blob/master/purl/url.py
        to test quickly, will probably need refactor in the future.

        Return or set a dictionary of query params
        :param dict value: new dictionary of values
        """
        if value is not None:
            return URL._mutate(self, query=unicode_urlencode(value, doseq=True))

        if qs_func is None:
            qs_func = original_parse_qs

        if query is None:
            query = '' if self._tuple.query is None else self._tuple.query

        # In Python 2.6, urlparse needs a bytestring so we encode and then
        # decode the result.
        if not six.PY3:
            result = qs_func(to_utf8(query), True)
            return dict_to_unicode(result)

        return qs_func(query, True)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-07-15 16:10
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('regex_match', '0006_auto_20160916_1428'),
    ]

    operations = [
        migrations.AlterField(
            model_name='parser',
            name='parser_method_params',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('regex_match', '0003_auto_20151110_1315'),
    ]

    operations = [
        migrations.AlterField(
            model_name='regexrule',
            name='regex',
            field=models.CharField(max_length=64),
        ),
    ]



import os
from inflection import camelize


def _import_models():
    exports = []
    package_path = os.path.dirname(__file__)
    glob_vars, loc_vars = globals(), locals()

    for module_file in os.listdir(package_path):
        module_file_name = module_file.split('.')

        if module_file_name[0][0] != '_' and len(module_file_name) > 1 and module_file_name[1] == 'py':
            module_name = module_file_name[0]
            subpackage = 'regex_match.objects.parsers'
            module = __import__(subpackage, glob_vars, loc_vars, [module_name])
            submodule = module.__dict__.get(module_name)
            model_name = camelize(module_name)
            model = submodule.__dict__.get(model_name)

            glob_vars.update({model_name: model})
            exports.append(model_name)

    return exports

if __name__ != '__main__':
    __all__ = _import_models()



from django.db import models

from .parser import Parser


class ExceptionParser(Parser):
    name = models.CharField(max_length=32)
    model_exception = models.ForeignKey(
        "ModelException", on_delete=models.CASCADE, related_name="parsers"
    )



from regex_match.objects.model_matcher import ModelMatcher
from regex_match.objects.parsers import UrlParser


class ExceptionParser(object):

    def __init__(self, model, model_field, url):
        self.url = url
        self.obj = UrlParser(url)
        self.match = ModelMatcher(
            model=model,
            model_field=model_field,
            obj=self.obj
        ).get()

    def is_matched(self):
        return self.match is not None

    def format_template(self):
        if self.match:
            return self.match.format_template(self.obj)
        else:
            return self.url



from funcy import keep

from regex_match.models import MatchingRule, ParserMethod, RegexRule
from regex_match.objects.parsers import UrlParser


class UrlToMatchingRule(object):
    default_regex_template_domain = "((^.*[.])|^){}$"
    default_regex_template_path = "^{}(/.*|)$"
    default_regex_template_query = "(^|.+&){}={}($|&.+)"

    def __init__(
        self,
        model_rule,
        urls,
        priority_func=None,
        description="",
        regex_template_domain=None,
        regex_template_path=None,
        regex_template_query=None,
    ):
        self.regex_template_domain = (
            regex_template_domain or self.default_regex_template_domain
        )
        self.regex_template_path = (
            regex_template_path or self.default_regex_template_path
        )
        self.regex_template_query = (
            regex_template_query or self.default_regex_template_query
        )

        self.model_rule = model_rule
        self.urls = urls
        self.priority_func = priority_func
        self.description = description

    def create(self):
        matching_rule_array = keep(self._create_matching_rule, self.urls)
        return list(matching_rule_array)

    def _create_matching_rule(self, url):
        obj = UrlParser(url)

        matching_rule = MatchingRule(
            model_rule=self.model_rule, description=self.description
        )
        matching_rule.save()

        domain_rule = self._create_domain_matching_rule(matching_rule, obj)

        path_rule = self._create_path_matching_rule(
            matching_rule, obj, bool(obj.url.query_params())
        )

        query_rules = self._create_query_matching_rule(matching_rule, obj)

        if self.priority_func is not None:
            self.priority = self.priority_func(domain_rule, path_rule, query_rules)

        return matching_rule

    def _create_domain_matching_rule(self, matching_rule, url):
        if url.domain() != "":
            parser_method = ParserMethod.objects.get(name="domain")
            return RegexRule.objects.create(
                matching_rule=matching_rule,
                parser_method=parser_method,
                regex=self.regex_template_domain.format(
                    url.domain().replace(".", "[.]")
                ),
            )

    def _create_path_matching_rule(self, matching_rule, url, has_params):
        if url.path() != "":
            regex_template = self.regex_template_path
            parser_method = ParserMethod.objects.get(name="path")
            return RegexRule.objects.create(
                matching_rule=matching_rule,
                parser_method=parser_method,
                regex=regex_template.format(url.path().replace(".", "[.]")),
            )

    def _create_query_matching_rule(self, matching_rule, url):
        result = list()
        query_params = url.url.query_params()
        query_params = {
            key: self.replace_spaces_in_query_values(value)
            for key, value in query_params.items()
        }
        for key, values in query_params.items():
            parser_method = ParserMethod.objects.get(name="query")
            rule = RegexRule.objects.create(
                matching_rule=matching_rule,
                parser_method=parser_method,
                regex=self.regex_template_query.format(
                    key, values[0].replace(".", "[.]").replace("+", "[+]")
                ),
            )
            result.append(rule)
        return result

    def replace_spaces_in_query_values(self, query_values):
        new_query_values = []
        for value_str in query_values:
            new_query_values.append(value_str.replace(" ", "+"))
        return new_query_values



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='MatchingRule',
            fields=[
                ('id', models.AutoField(serialize=False, verbose_name='ID', auto_created=True, primary_key=True)),
                ('description', models.CharField(max_length=32)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name='ModelRule',
            fields=[
                ('id', models.AutoField(serialize=False, verbose_name='ID', auto_created=True, primary_key=True)),
                ('model_field', models.CharField(max_length=16)),
                ('model_field_type', models.CharField(max_length=16)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Parser',
            fields=[
                ('id', models.AutoField(serialize=False, verbose_name='ID', auto_created=True, primary_key=True)),
                ('parser_method_params', models.CharField(blank=True, max_length=16, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ParserMethod',
            fields=[
                ('id', models.AutoField(serialize=False, verbose_name='ID', auto_created=True, primary_key=True)),
                ('type', models.CharField(max_length=50)),
                ('name', models.CharField(max_length=50)),
            ],
        ),
        migrations.CreateModel(
            name='ExceptionParser',
            fields=[
                ('parser_ptr', models.OneToOneField(primary_key=True, parent_link=True, to='regex_match.Parser', auto_created=True, serialize=False)),
                ('name', models.CharField(max_length=32)),
            ],
            options={
                'abstract': False,
            },
            bases=('regex_match.parser',),
        ),
        migrations.CreateModel(
            name='ModelException',
            fields=[
                ('modelrule_ptr', models.OneToOneField(primary_key=True, parent_link=True, to='regex_match.ModelRule', auto_created=True, serialize=False)),
                ('name', models.CharField(max_length=32)),
                ('template', models.CharField(max_length=64)),
            ],
            options={
                'abstract': False,
            },
            bases=('regex_match.modelrule',),
        ),
        migrations.CreateModel(
            name='RegexRule',
            fields=[
                ('parser_ptr', models.OneToOneField(primary_key=True, parent_link=True, to='regex_match.Parser', auto_created=True, serialize=False)),
                ('regex', models.CharField(max_length=32)),
            ],
            options={
                'abstract': False,
            },
            bases=('regex_match.parser',),
        ),
        migrations.AlterUniqueTogether(
            name='parsermethod',
            unique_together=set([('name', 'type')]),
        ),
        migrations.AddField(
            model_name='parser',
            name='parser_method',
            field=models.ForeignKey(related_name='parsers', to='regex_match.ParserMethod'),
        ),
        migrations.AddField(
            model_name='parser',
            name='polymorphic_ctype',
            field=models.ForeignKey(related_name='polymorphic_regex_match.parser_set+', editable=False, to='contenttypes.ContentType', null=True),
        ),
        migrations.AddField(
            model_name='modelrule',
            name='polymorphic_ctype',
            field=models.ForeignKey(related_name='polymorphic_regex_match.modelrule_set+', editable=False, to='contenttypes.ContentType', null=True),
        ),
        migrations.AddField(
            model_name='matchingrule',
            name='model_rule',
            field=models.ForeignKey(related_name='matching_rules', to='regex_match.ModelRule'),
        ),
        migrations.AddField(
            model_name='regexrule',
            name='matching_rule',
            field=models.ForeignKey(related_name='regex_rules', to='regex_match.MatchingRule'),
        ),
        migrations.AddField(
            model_name='exceptionparser',
            name='model_exception',
            field=models.ForeignKey(related_name='parsers', to='regex_match.ModelException'),
        ),
    ]



from django.db import models


class ParserMethod(models.Model):
    type = models.CharField(max_length=50)
    name = models.CharField(max_length=100)

    def __str__(self):
        return '{}.{}()'.format(self.type, self.name)

    class Meta:
        unique_together = ('name', 'type')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re

from setuptools import setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, "__init__.py")).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [
        dirpath
        for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]


version = get_version("regex_match")

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))
install_requires = [
    "inflection>=0.3.0",
    "django-polymorphic==1.3",
    "funcy==1.4",
    "purl==1.0.3",
    "toolz==0.12.0",
]

setup(
    name="django-regex-match",
    version=version,
    packages=get_packages("regex_match"),
    include_package_data=True,
    license="",  # example license
    description="",
    url="http://www.admetricks.com/",
    author="Dev Admx",
    author_email="dev@admetricks.com",
    zip_safe=False,
    install_requires=install_requires,
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",  # example license
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        # Replace these appropriately if you are stuck on Python 2.
    ],
)



from django.db import models
from polymorphic.models import PolymorphicModel


class Parser(PolymorphicModel):
    parser_method = models.ForeignKey(
        "ParserMethod", on_delete=models.CASCADE, related_name="parsers"
    )
    parser_method_params = models.CharField(max_length=100, blank=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True, editable=False)
    updated_at = models.DateTimeField(auto_now=True, editable=False)

    def method(self):
        # Add a cache system here
        return self.parser_method.name

    def params(self):
        return self.parser_method_params

    def parse_obj(self, obj):
        method = self.method()
        params = self.params()
        parser_method = getattr(obj, method)

        if params:
            return parser_method(params)
        else:
            return parser_method()
